/* $Id: main.c 489 2013-07-16 06:04:45Z michi $
 * $URL: https://svn.trisoft.at/AVR/PBX/main.c $
 *
 * PBX Personal Branch Exchange
 *
 * Copyright (C) 2013 Michael Reinelt <michael@reinelt.co.at>
 *
 */

#include <stdint.h>
#include <stdlib.h>
#include <stdarg.h>

#include <avr/io.h>
#include <avr/interrupt.h>
#include <avr/sleep.h>
#include <util/delay.h>
#include <util/atomic.h>
#define BAUD 57600
#include <util/setbaud.h>

#include "qprintf.h"


#define PBX_LINES 5

#define NUMBER_1 0x666		/* Papa Keller */
#define NUMBER_2 0x144		/* Mama Keller */
#define NUMBER_3 0x111		/* Wohnzimmer */
#define NUMBER_4 0x133		/* Benni */
#define NUMBER_5 0x122		/* David */


// Debugging
#define HISTOGRAM


// Rufspannung
#define RNG_PORT PORTB
#define RNG_BIT1 _BV(1)
#define RNG_BIT2 _BV(2)


// Relais
#define K1_PORT PORTB
#define K1_BIT _BV(0)

#define K2_PORT PORTD
#define K2_BIT _BV(7)

#define K3_PORT PORTD
#define K3_BIT _BV(6)

#define K4_PORT PORTD
#define K4_BIT _BV(4)

#define K5_PORT PORTD
#define K5_BIT _BV(2)


// Hörtöne
#define T1_PORT PORTD
#define T1_BIT _BV(3)

#define T2_PORT PORTD
#define T2_BIT _BV(5)

// Umrechnung Frequenz in CTC-Wert
#define FREQ2CTC(f) (uint8_t)((double)F_CPU / 256 / (f) / 2 + 0.5)


// Timings
#define T_RSM 625
#define T_ADC 85


// ADC
#define ADC_PORT PORTC
#define ADC_DDR DDRC
#define ADC_BIT0 0
#define ADC_BIT1 1
#define ADC_BIT2 2
#define ADC_BIT3 3
#define ADC_BIT4 4

// Hysteresis
#define ADC_THRESHOLD_LO 45
#define ADC_THRESHOLD_HI 55


// Ruf- und Signalmaschine
typedef enum { TONE_NONE, TONE_DIAL, TONE_RING, TONE_BUSY, TONE_FAIL } tone_t;
typedef struct {
    uint8_t bell:1;		// Wecker
    tone_t tone1:3;		// Hörton Sprechschiene
    tone_t tone2:3;		// Hörton Warteschiene
} RSM_t;

static volatile RSM_t RSM;


// Telefonanlage
typedef enum { STATE_IDLE, STATE_WAITING, STATE_DIALING, STATE_FAIL, STATE_BUSY, STATE_RINGING, STATE_TALKING, STATE_ERROR } state_t;
typedef struct {
    state_t State;		// FSM
    uint8_t Caller;		// calling line
    uint8_t Callee;		// called line
    uint8_t Relay;		// Relay bit field
} PBX_t;

PBX_t PBX;


// Status Telefonleitungen
static volatile uint8_t PBX_Hook;
static volatile uint16_t PBX_Dial;


// Histogramm Wahlimpulse
#ifdef HISTOGRAM
uint16_t Histo1;
uint16_t Histo2;
uint16_t Histo3;
#endif


// UART 
static volatile uint8_t UART_Buffer[256];
static volatile uint8_t UART_TxHead;
static volatile uint8_t UART_TxTail;

void uart_init(void)
{
    UART_TxHead = 0;
    UART_TxTail = 0;

    // AVR USART
    UBRR0H = UBRRH_VALUE;	// bit rate high 
    UBRR0L = UBRRL_VALUE;	// bit rate low
#if USE_2X
    UCSR0A |= (1 << U2X0);
#else
    UCSR0A &= ~(1 << U2X0);
#endif
    UCSR0B |= _BV(TXEN0);	// enable TX
    UCSR0C = _BV(UCSZ00) | _BV(UCSZ01);	// async 8N1     
}


ISR(USART_UDRE_vect)
{
    if (UART_TxHead != UART_TxTail) {
	UDR0 = UART_Buffer[++UART_TxTail];	// send next byte
    } else {
	UCSR0B &= ~_BV(UDRIE0);	// finished, disable interrupt
    }
}


static void uart_put(const uint8_t data)
{
    uint8_t tmpHead = UART_TxHead + 1;	// pointer to next free character 
    while (tmpHead == UART_TxTail);	// wait for free space in the buffer
    UART_Buffer[tmpHead] = data;	// put character into buffer
    UART_TxHead = tmpHead;	// set write pointer
    UCSR0B |= _BV(UDRIE0);	// enable interrupt
}


static void uart_printf(const char *format, ...)
{
    // prepare string
    va_list ap;
    static char buffer[255];
    va_start(ap, format);
    uint8_t len = vqprintf(buffer, sizeof(buffer), format, ap);
    va_end(ap);

    // send string 
    for (uint8_t i = 0; i < len; i++) {
	uart_put(buffer[i]);
    }
}


static void timer_init(void)
{

    // Timer 0: Ton2 = OC0B = Hörton
    // 16 MHz, Prescaler 256
    TCCR0A = (0 << WGM02) | (1 << WGM01) | (0 << WGM00) | (1 << COM0B1) | (0 << COM0B0);	// Mode 2: CTC; Clear OC0B on Compare Match
    TCCR0B = (1 << CS02) | (0 << CS01) | (0 << CS00);	// Prescaler 256 
    OCR0A = 0;			// used for CTC
    OCR0B = 0;			// used to toggle OC0B
    TIMSK0 = 0;			// no interrupts


    // Timer 1: RSM 100 Hz, ADC 1000 Hz
    // 16 MHz, Prescaler 256
    TCCR1A = (0 << COM1A1) | (0 << COM1A0) | (0 << COM1B1) | (0 << COM1B0) | (0 << WGM11) | (0 << WGM10);	// Mode 0: normal operation
    TCCR1B = (0 << ICNC1) | (0 << ICES1) | (0 << WGM13) | (0 << WGM12) | (1 << CS12) | (0 << CS11) | (0 << CS10);	// Prescaler 256
    TCCR1C = (0 << FOC1A) | (0 << FOC1B);	// normal operation
    OCR1A = T_RSM;		// RSM 100 Hz
    OCR1B = T_ADC;		// ADC 1000 Hz
    TIMSK1 = (0 << ICIE1) | (1 << OCIE1B) | (1 << OCIE1A) | (0 << TOIE1);	// Output Compare A/B Match Interrupt Enable


    // Timer 2: Ton1 = OC2B = Hörton
    // 16 MHz, Prescaler 256
    TCCR2A = (0 << WGM22) | (1 << WGM21) | (0 << WGM20) | (1 << COM2B1) | (0 << COM2B0);	// Mode 2: CTC; Clear OC2B on Compare Match
    TCCR2B = (1 << CS22) | (1 << CS21) | (0 << CS20);	// Prescaler 256 
    OCR2A = 0;			// used for CTC
    OCR2B = 0;			// used to toggle OC2B
    TIMSK2 = 0;			// no interrupts
}




// Timer 1 compare match A handler
// RSM working horse
ISR(TIMER1_COMPA_vect)
{
    OCR1A += T_RSM;		// setup next event

    RSM_t rsm = RSM;		// fetch volatile data into local variable

    // Klingeln

    static uint8_t wave = 0;	// ringer wave (interleave)
    static uint16_t tick = 0;	// ringer tick (100 Hz)

    if (rsm.bell) {
	wave = (tick < 30 || (tick > 130 && tick < 230)) ? (wave + 1) & 3 : 0;
	if (++tick >= 630) {
	    tick = 130;
	}
    } else {
	tick = 0;
	wave = 0;
    }
    switch (wave) {
    case 1:
	RNG_PORT |= RNG_BIT1;
	RNG_PORT &= ~RNG_BIT2;
	break;
    case 3:
	RNG_PORT &= ~RNG_BIT1;
	RNG_PORT |= RNG_BIT2;
	break;
    default:			// catches 0 and 2
	RNG_PORT &= ~RNG_BIT1;
	RNG_PORT &= ~RNG_BIT2;
    }


    // Ton 1: Klingel/Sprechschiene

    uint8_t freq1;		// Hörton Frequenz (eigentlich CTC-Wert)
    static uint16_t tick1 = 0;	// tone1 tick (100 Hz)

    switch (rsm.tone1) {
    case TONE_DIAL:
	// Wählton: 425 Hz, Dauerton
	freq1 = FREQ2CTC(425);
	break;
    case TONE_RING:
	// Freiton: 425 Hz, 1000 msec Ton, 4000 msec Pause
	freq1 = (tick1 < 30 || (tick1 > 130 && tick1 < 230)) ? FREQ2CTC(425) : 0;
	if (++tick1 >= 630) {
	    tick1 = 130;
	}
	break;
    case TONE_BUSY:
	// Besetztton: 425 Hz, 480 msec Ton, 480 msec Pause
	freq1 = (tick1 < 48) ? FREQ2CTC(425) : 0;
	if (++tick1 >= 96) {
	    tick1 = 0;
	}
	break;
    case TONE_FAIL:
	// Hinweiston: 950/1400/1800 Hz, je 330 msec, 1000 msec Pause 
	// "Kein Anschluss unter dieser Nummer"
	if (tick1 < 33) {
	    freq1 = FREQ2CTC(950);
	} else if (tick1 < 66) {
	    freq1 = FREQ2CTC(1400);
	} else if (tick1 < 99) {
	    freq1 = FREQ2CTC(1800);
	} else {
	    freq1 = 0;
	}
	if (++tick1 >= 199) {
	    tick1 = 0;
	}
	break;
    case TONE_NONE:
    default:
	freq1 = 0;
	tick1 = 0;
    }

    if (freq1 > 0) {
	OCR0A = freq1;		// CTC
	OCR0B = freq1;		// toggle OC0B
	TCCR0A = (0 << WGM02) | (1 << WGM01) | (0 << WGM00) | (0 << COM0B1) | (1 << COM0B0);	// Toggle OC0B on Compare Match
    } else {
	TCCR0A = (0 << WGM02) | (1 << WGM01) | (0 << WGM00) | (1 << COM0B1) | (0 << COM0B0);	// Clear OC0B on Compare Match
    }


    // Ton 2: Warte-Schiene

    uint8_t freq2;		// Hörton Frequenz (eigentlich CTC-Wert)
    static uint16_t tick2 = 0;	// tone2 tick (100 Hz)

    switch (rsm.tone2) {
    case TONE_DIAL:
	// Wählton: 425 Hz, Dauerton
	freq2 = FREQ2CTC(425);
	break;
    case TONE_RING:
	// Freiton: 425 Hz, 1000 msec Ton, 4000 msec Pause
	freq2 = (tick2 < 30 || (tick2 > 130 && tick2 < 230)) ? FREQ2CTC(425) : 0;
	if (++tick2 >= 630) {
	    tick2 = 130;
	}
	break;
    case TONE_BUSY:
	// Besetztton: 425 Hz, 480 msec Ton, 480 msec Pause
	freq2 = (tick2 < 48) ? FREQ2CTC(425) : 0;
	if (++tick2 >= 96) {
	    tick2 = 0;
	}
	break;
    case TONE_FAIL:
	// Hinweiston: 950/1400/1800 Hz, je 330 msec, 1000 msec Pause 
	// "Kein Anschluss unter dieser Nummer"
	if (tick2 < 33) {
	    freq2 = FREQ2CTC(950);
	} else if (tick2 < 66) {
	    freq2 = FREQ2CTC(1400);
	} else if (tick2 < 99) {
	    freq2 = FREQ2CTC(1800);
	} else {
	    freq2 = 0;
	}
	if (++tick2 >= 199) {
	    tick2 = 0;
	}
	break;
    case TONE_NONE:
    default:
	freq2 = 0;
	tick2 = 0;
    }

    if (freq2 > 0) {
	OCR2A = freq2;		// CTC
	OCR2B = freq2;		// toggle OC2B
	TCCR2A = (0 << WGM22) | (1 << WGM21) | (0 << WGM20) | (0 << COM2B1) | (1 << COM2B0);	// Toggle OC2B on Compare Match
    } else {
	TCCR2A = (0 << WGM22) | (1 << WGM21) | (0 << WGM20) | (1 << COM2B1) | (0 << COM2B0);	// Clear OC2B on Compare Match
    }
}


// Timer 1 compare match B handler
// ADC trigger
ISR(TIMER1_COMPB_vect)
{
    OCR1B += T_ADC;		// setup next event

    // start ADC conversion
    ADCSRA |= (1 << ADSC);
}


void port_init(void)
{
    PORTB = 0;			// all off
    DDRB = RNG_BIT1 | RNG_BIT2 | K1_BIT;

    PORTD = 0;			// all off
    DDRD = K2_BIT | K3_BIT | K4_BIT | K5_BIT | T1_BIT | T2_BIT;
}


static void adc_init(void)
{
    // setup port
    ADC_DDR &= ~((1 << ADC_BIT0) || (1 << ADC_BIT1) || (1 << ADC_BIT2) || (1 << ADC_BIT3) || (1 << ADC_BIT4));
    ADC_PORT &= ~((1 << ADC_BIT0) || (1 << ADC_BIT1) || (1 << ADC_BIT2) || (1 << ADC_BIT3) || (1 << ADC_BIT4));

    // disable digital inputs
    DIDR0 |= (1 << ADC_BIT0) || (1 << ADC_BIT1) || (1 << ADC_BIT2) || (1 << ADC_BIT3) || (1 << ADC_BIT4);

    // setup ADC
    ADMUX = (1 << REFS1) | (1 << REFS0) | (1 << ADLAR) | (0 << MUX3) | (0 << MUX2) | (0 << MUX1) | (0 << MUX0);	// internal 1.1V reference; left-adjust result; select ADC0
    ADCSRA = (1 << ADEN) | (0 << ADSC) | (0 << ADATE) | (1 << ADIE) | (1 << ADPS2) | (1 << ADPS1) | (1 << ADPS0);	// eanble ADC, prescaler 128
}


// ADC ISR
ISR(ADC_vect)
{
    static uint8_t dials[PBX_LINES];
    static uint8_t ticks[PBX_LINES];
    static uint8_t fast_bits[PBX_LINES];
    static uint8_t slow_bits[PBX_LINES];

    uint8_t line = ADMUX & 0x07;	// 3 bits 
    uint8_t volt = ADCH;	// read upper 8 bits

    uint8_t dial = dials[line];
    uint8_t tick = ticks[line];
    uint8_t fast_bit = fast_bits[line];
    uint8_t slow_bit = slow_bits[line];

    uint8_t threshold = (fast_bit & 1) ? ADC_THRESHOLD_LO : ADC_THRESHOLD_HI;
    fast_bit <<= 1;
    if (volt > threshold)
	fast_bit |= 1;

    // debounce: fast_bit has not changed three times
    if ((fast_bit & 7) == 0) {
	slow_bit &= ~1;
    } else if ((fast_bit & 7) == 7) {
	slow_bit |= 1;
    }
    // slow bits
    if (++tick > 2) {
	tick = 0;

	// detect dial pulse
	// pattern: xx100001 or xxx10001 or xxxx1001
	// mask:    00111111    00011111    00001111
	// hex:     21/3f       11/1f       09/0f
	if ((slow_bit & 0x3f) == 0x21 || (slow_bit & 0x1f) == 0x11 || (slow_bit & 0x0f) == 0x09) {
	    if (dial == 0) {
		PBX_Dial = (PBX_Dial << 4) | 0x0f;	// dial in progress
	    }
	    dial++;
#ifdef HISTOGRAM
	    if ((slow_bit & 0x3f) == 0x21) {
		Histo1++;
	    } else if ((slow_bit & 0x1f) == 0x11) {
		Histo2++;
	    } else if ((slow_bit & 0x0f) == 0x09) {
		Histo3++;
	    }
#endif
	}
	// all bits high: phone engaged, dial pause
	else if (slow_bit == 0xff) {
	    PBX_Hook |= (1 << line);
	    if (dial > 0) {
		PBX_Dial = (PBX_Dial & 0xfff0) | dial;
		dial = 0;
	    }
	}
	// all bits low: phone hung up
	else if (slow_bit == 0) {
	    dial = 0;
	    PBX_Hook &= ~(1 << line);
	}
	// shift and duplicate LSB
	slow_bit = __builtin_avr_insert_bits(0x65432100, slow_bit, 0);
    }

    dials[line] = dial;
    ticks[line] = tick;
    fast_bits[line] = fast_bit;
    slow_bits[line] = slow_bit;

    // switch to next line
    line = (line < PBX_LINES - 1) ? line + 1 : 0;
    ADMUX = __builtin_avr_insert_bits(0xffff3210, line, ADMUX);
}


// Counting bits set, Brian Kernighan's way
static uint8_t popcount(uint8_t v)
{
    uint8_t c;
    for (c = 0; v != 0; c++) {
	v &= v - 1;		// clear the least significant bit set
    }
    return c;
}


static uint8_t selector(uint16_t number)
{
    switch (number) {
    case NUMBER_1:
	return 0x01;
    case NUMBER_2:
	return 0x02;
    case NUMBER_3:
	return 0x04;
    case NUMBER_4:
	return 0x08;
    case NUMBER_5:
	return 0x10;
    }
    return 0;
}


int main(void)
{

    // power saving stuff
    ACSR &= _BV(ACD);		// disable analog comparator

    port_init();		// setup I/O ports
    adc_init();			// setup ADC
    uart_init();		// start UART
    timer_init();		// start timers

    // init state machine
    PBX.State = STATE_IDLE;
    PBX.Caller = 0;
    PBX.Callee = 0;
    PBX.Relay = 0;

    RSM.bell = 0;
    RSM.tone1 = TONE_NONE;
    RSM.tone2 = TONE_NONE;

    sei();			// now enable interrupts

    uart_printf("Hello, World!\n");


    uint16_t old_dial = 0;
    uint8_t old_state = 255;
    uint8_t old_relay = 0;

    while (1) {

	// fetch voltile stuff into local memory
	uint8_t Hook = PBX_Hook;
	uint16_t Dial = PBX_Dial;

	// FSM
	switch (PBX.State) {

	case STATE_IDLE:
	    PBX_Dial = 0;
	    PBX.Caller = 0;
	    PBX.Callee = 0;
	    PBX.Relay = 0;
	    RSM.bell = 0;
	    RSM.tone1 = TONE_NONE;
	    RSM.tone2 = TONE_NONE;
	    uint8_t pop = popcount(Hook);	// number of active lines
	    if (pop == 1) {
		// one phone went online => wait for dialing
		PBX.Caller = Hook;
		PBX.State = STATE_WAITING;
	    } else if (pop > 1) {
		// two or more phones went online
		PBX.State = STATE_ERROR;
	    }
	    break;

	case STATE_WAITING:
	    PBX.Callee = 0;
	    PBX.Relay = 0;
	    RSM.bell = 0;
	    RSM.tone1 = TONE_DIAL;
	    RSM.tone2 = TONE_NONE;
	    if (Hook == 0) {
		// all phones inactive => idle
		PBX.State = STATE_IDLE;
	    } else if (Hook ^ PBX.Caller) {
		// any other phone went online
		PBX.State = STATE_ERROR;
	    } else if (Dial > 0) {
		// a number has been dialed
		PBX.State = STATE_DIALING;
	    }
	    break;

	case STATE_DIALING:
	    PBX.Callee = 0;
	    PBX.Relay = 0;
	    RSM.bell = 0;
	    RSM.tone1 = TONE_NONE;
	    RSM.tone2 = TONE_NONE;
	    if (Hook == 0) {
		// all phones inactive => idle
		PBX.State = STATE_IDLE;
	    } else if (Hook ^ PBX.Caller) {
		// any other phone went online
		PBX.State = STATE_ERROR;
	    } else if (Dial > 0x00ff && ((uint8_t) Dial & 0x0f) != 0x0f) {
		// three digits have been dialed
		uint8_t callee = selector(Dial);
		if (callee == 0) {
		    // kein Anschluss unter dieser Nummer
		    PBX.State = STATE_FAIL;
		} else if (callee == PBX.Caller) {
		    // calling myself
		    PBX.State = STATE_BUSY;
		} else {
		    PBX.Callee = callee & ~PBX.Caller;	// do not call ourselves 
		    PBX.State = STATE_RINGING;
		}
	    }
	    break;

	case STATE_FAIL:
	    PBX.Relay = 0;
	    RSM.bell = 0;
	    RSM.tone1 = TONE_FAIL;
	    RSM.tone2 = TONE_NONE;
	    if (PBX_Hook == 0) {
		// all phones inactive again
		PBX.State = STATE_IDLE;
	    } else if (Hook ^ PBX.Caller) {
		// any other phone went online
		PBX.State = STATE_ERROR;
	    }
	    break;

	case STATE_BUSY:
	    PBX.Relay = 0;
	    RSM.bell = 0;
	    RSM.tone1 = TONE_BUSY;
	    RSM.tone2 = TONE_NONE;
	    if (PBX_Hook == 0) {
		// all phones inactive again
		PBX.State = STATE_IDLE;
	    } else if (Hook ^ PBX.Caller) {
		// any other phone went online
		PBX.State = STATE_ERROR;
	    }
	    break;

	case STATE_RINGING:
	    PBX.Relay = PBX.Callee;
	    RSM.bell = 1;
	    RSM.tone1 = TONE_RING;
	    RSM.tone2 = TONE_NONE;
	    if (Hook == 0) {
		// all phones inactive => idle
		PBX.State = STATE_IDLE;
	    } else if (Hook == (PBX.Caller | PBX.Callee)) {
		// Callee has answered
		PBX.State = STATE_TALKING;
	    } else if (Hook ^ PBX.Caller) {
		// any other phone went online
		PBX.State = STATE_ERROR;
	    }
	    break;

	case STATE_TALKING:
	    PBX.Relay = PBX.Caller | PBX.Callee;
	    RSM.bell = 0;
	    RSM.tone1 = (Hook & ~(PBX.Caller | PBX.Callee)) ? TONE_BUSY : TONE_NONE;	// other phones get a busy signal
	    RSM.tone2 = TONE_NONE;
	    if (!(Hook & PBX.Caller && Hook & PBX.Callee)) {
		// Caller or Callee left the call
		PBX.State = STATE_ERROR;
	    }
	    break;

	case STATE_ERROR:
	default:
	    PBX.Caller = 0;
	    PBX.Callee = 0;
	    PBX.Relay = 0;
	    RSM.bell = 0;
	    RSM.tone1 = TONE_BUSY;
	    RSM.tone2 = TONE_NONE;
	    if (PBX_Hook == 0) {
		// all phones inactive again
		PBX.State = STATE_IDLE;
	    }
	    break;
	}


	// Relais
	if (PBX.Relay & 0x10) {
	    K1_PORT |= K1_BIT;
	} else {
	    K1_PORT &= ~K1_BIT;
	}
	if (PBX.Relay & 0x08) {
	    K2_PORT |= K2_BIT;
	} else {
	    K2_PORT &= ~K2_BIT;
	}
	if (PBX.Relay & 0x04) {
	    K3_PORT |= K3_BIT;
	} else {
	    K3_PORT &= ~K3_BIT;
	}
	if (PBX.Relay & 0x02) {
	    K4_PORT |= K4_BIT;
	} else {
	    K4_PORT &= ~K4_BIT;
	}
	if (PBX.Relay & 0x01) {
	    K5_PORT |= K5_BIT;
	} else {
	    K5_PORT &= ~K5_BIT;
	}

	if (old_state != PBX.State || old_dial != Dial || old_relay != PBX.Relay) {
	    uart_printf("State: %d => %d Hook %08b Relay %08b Caller:%d Dial:%x Callee:%d Histo %d %d %d\n", old_state, PBX.State, Hook, PBX.Relay, PBX.Caller, Dial, PBX.Callee,
			Histo1, Histo2, Histo3);
	    old_state = PBX.State;
	    old_dial = Dial;
	    old_relay = PBX.Relay;
	}

    }
}
