/* $Id: qprintf.c 475 2013-07-07 04:26:10Z michi $
 * $URL: https://svn.trisoft.at/AVR/PBX/qprintf.c $
 *
 * simple but quick snprintf() replacement
 *
 * Copyright (C) 2012 Michael Reinelt <michael@reinelt.co.at>
 *
 */

/* 
 * exported functions:
 * 
 * uint8_t qprintf(char *str, uint8_t size, const char *format, ...)
 *   works like snprintf(), but format only knows about %d, %u, %f and %s
 *     and for the numbers an optional length like %<len>d. If <len> beginns
 *     with '0' the free space is filled with '0's, otherwise with ' '
 */


#include <stdlib.h>
#include <stdint.h>
#include <stdarg.h>
#include <string.h>
#include <math.h>

/* from ftoa_engine.h */
#define	FTOA_MINUS	1
#define	FTOA_ZERO	2
#define	FTOA_INF	4
#define	FTOA_NAN	8
#define	FTOA_CARRY	16

int __ftoa_engine(double val, char *buf, unsigned char prec, unsigned char maxdgs);



static char *my_itoa(char *buffer, const uint8_t size, const int32_t value, const uint8_t width, const uint8_t fill0)
{
    /* remember sign of value */
    uint32_t v;
    uint8_t sign;
    if (value < 0) {
	v = -value;
	sign = 1;
    } else {
	v = value;
	sign = 0;
    }

    char *p;
    p = buffer + size - 1;	/* p points to last char */
    *p = '\0';			/* set terminating zero */

    uint8_t l = 0;
    do {
	*--p = (v % 10) + '0';
	v /= 10;
	l++;
    } while (v);

    if (sign && !fill0) {
	*--p = '-';
	l++;
	sign = 0;
    }

    /* overflow? */
    if (width && l > width) {
	p = buffer + size - 1;
	l = width;
	do {
	    *--p = '#';
	} while (--l);
	return p;
    }

    /* pad to width */
    char pad = fill0 ? '0' : ' ';
    while (l++ < width - sign && p > buffer) {
	*--p = pad;
    }
    if (sign)
	*--p = '-';

    return p;
}


static char *my_utoa(char *buffer, const uint8_t size, const uint32_t value, const uint8_t base, const uint8_t width, const uint8_t fill0)
{
    uint32_t v = value;

    char *p;
    p = buffer + size - 1;	/* p points to last char */
    *p = '\0';			/* set terminating zero */

    uint8_t l = 0;
    do {
	uint8_t digit = v % base;
	*--p = (digit < 10) ? digit + '0' : digit - 10 + 'a';
	v = v / base;
	l++;
    } while (v);

    /* overflow? */
    if (width && l > width) {
	p = buffer + size - 1;
	l = width;
	do {
	    *--p = '#';
	} while (--l);
	return p;
    }

    /* pad to width */
    char pad = fill0 ? '0' : ' ';
    while (l++ < width && p > buffer) {
	*--p = pad;
    }

    return p;
}


static char *my_ftoa(char *buffer, const uint8_t size, const double value, const uint8_t width, const uint8_t prec, const uint8_t fill0)
{
    /* width is the total (maximum) field width */
    /* if width is zero, resulting string grows as needed up to <size> */
    /* prec is limited to 12 by caller */
    /* INF => ###.##, NAN => ---.-- according to width and prec */
    /* if output does not fit into width, it will be displayed as ###.## */
    /* we DO NOT reduce precision because of rounding issues */


    char result[9];
    int8_t exp = __ftoa_engine(value, result, 7, prec + 1);
    uint8_t flag = result[0];
    uint8_t sign = flag & FTOA_MINUS ? 1 : 0;

    int8_t len = (exp > 0 ? exp + 1 : 1);
    if (sign)
	len += 1;
    if (prec)
	len += prec + 1;

    if ((width && len > width) || len > size - 1)
	flag |= FTOA_INF;

    if (flag & (FTOA_NAN | FTOA_INF)) {
	char *p = buffer + size - 1;
	*p = '\0';
	char pad = (flag & FTOA_NAN) ? '-' : '#';
	uint8_t l = width ? width : prec ? prec + 2 : 1;
	uint8_t d = prec;
	while (l--) {
	    *--p = d-- ? pad : '.';
	}
	return p;
    }


    char *p = buffer;

    int8_t pad = width > len ? width - len : 0;
    if (!fill0) {
	while (pad) {
	    *p++ = ' ';
	    pad--;
	}
    }

    if (sign)
	*p++ = '-';

    while (pad) {
	*p++ = '0';
	pad--;
    }

    int8_t ndigs = prec + 1 + exp;
    if ((flag & FTOA_CARRY) && result[1] == '1')
	ndigs--;
    if (ndigs < 1)
	ndigs = 1;
    else if (ndigs > 8)
	ndigs = 8;

    char c;
    int8_t n = exp > 0 ? exp : 0;
    do {
	if (n == -1)
	    *p++ = '.';
	c = (n <= exp && n > exp - ndigs) ? result[exp - n + 1] : '0';
	if (--n < -prec)
	    break;
	*p++ = c;
    } while (1);
    if (n == exp && (result[1] > '5' || (result[1] == '5' && !(flag & FTOA_CARRY)))) {
	c = '1';
    }
    *p++ = c;
    *p = '\0';

    return buffer;
}


uint8_t vqprintf(char *str, const uint8_t size, const char *format, va_list ap)
{

    const char *src;
    char *dst;
    uint8_t len;
    uint8_t max;

    src = format;
    dst = str;
    len = 0;
    max = size - 1;		/* leave room for terminating zero */

    while (len < max) {

	if (*src == '%') {
	    char buf[17], *s;
	    uint8_t width = 0;
	    uint8_t prec = 2;
	    uint8_t fill0 = 0;
	    if (*++src == '0')
		fill0 = 1;
	    while (*src >= '0' && *src <= '9')
		width = width * 10 + (*src++ - '0');
	    if (width > 16)
		width = 16;
	    if (*src == '.') {
		prec = 0;
		src++;
		while (*src >= '0' && *src <= '9')
		    prec = prec * 10 + (*src++ - '0');
		if (prec > 12)
		    prec = 12;
	    }
	    switch (*src) {
	    case 's':
		src++;
		s = va_arg(ap, char *);
		while (len < max && *s != '\0') {
		    len++;
		    *dst++ = *s++;
		}
		break;
	    case 'd':
		src++;
		int16_t d = va_arg(ap, int16_t);
		s = my_itoa(buf, sizeof(buf), d, width, fill0);
		while (len < max && *s != '\0') {
		    len++;
		    *dst++ = *s++;
		}
		break;
	    case 'l':
		src++;
		int32_t l = va_arg(ap, int32_t);
		s = my_itoa(buf, sizeof(buf), l, width, fill0);
		while (len < max && *s != '\0') {
		    len++;
		    *dst++ = *s++;
		}
		break;
	    case 'b':
		src++;
		uint16_t b = va_arg(ap, uint16_t);
		s = my_utoa(buf, sizeof(buf), b, 2, width, fill0);
		while (len < max && *s != '\0') {
		    len++;
		    *dst++ = *s++;
		}
		break;
	    case 'x':
		src++;
		uint16_t x = va_arg(ap, uint16_t);
		s = my_utoa(buf, sizeof(buf), x, 16, width, fill0);
		while (len < max && *s != '\0') {
		    len++;
		    *dst++ = *s++;
		}
		break;
	    case 'u':
		src++;
		uint32_t u = va_arg(ap, uint32_t);
		s = my_utoa(buf, sizeof(buf), u, 10, width, fill0);
		while (len < max && *s != '\0') {
		    len++;
		    *dst++ = *s++;
		}
		break;
	    case 'f':
		src++;
		double f = va_arg(ap, double);
		s = my_ftoa(buf, sizeof(buf), f, width, prec, fill0);
		while (len < max && *s != '\0') {
		    len++;
		    *dst++ = *s++;
		}
		break;
		// Fixme
		/* uint8_t width = 1; */
		/* if (width > 0) */
		/*     width = width; */
		/* if (width > sizeof(buf)) */
		/*     width = sizeof(buf); */
		/* if (isnan(f) || isinf(f)) { */
		/*     for (uint8_t i = 0; i < width; i++) */
		/*      buf[i] = '-'; */
		/*     buf[width - prec - 1] = '.'; */
		/*     buf[width] = 0;  // terminate */
		/*     s = buf; */
		/* } else { */
		/*     s = dtostrf(f, width, prec, buf); */
		/* } */

	    default:
		len++;
		*dst++ = '%';
	    }
	} else {
	    len++;
	    *dst++ = *src;
	    if (*src++ == '\0')
		break;
	}
    }

    /* enforce terminating zero */
    if (len >= max && *(dst - 1) != '\0') {
	len++;
	*dst = '\0';
    }

    /* do not count terminating zero */
    return len - 1;
}

uint8_t qprintf(char *str, const uint8_t size, const char *format, ...)
{
    va_list ap;
    va_start(ap, format);
    uint8_t len = vqprintf(str, size, format, ap);
    va_end(ap);
    return len;
}
